class _{{class_name}}(klass, concrete_class=True):
    __slots__ = {{slots}}
    # Copy any custom __eq__, etc into the concrete class
    # so that we can allow the user to call super()
    # to defer to the autogenerated __eq__, etc
    {% for key in data_class_attrs %}
    {{key}} = _dataclass_attrs["{{key}}"]
    {% endfor %}